/*
	TuxPuck - audio.cpp
	Copyright (C) 2001 Jacob Kroon

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

	Jacob Kroon - d00jkr@efd.lth.se
*/

/* includes */
#include <stdlib.h>
#include <string.h>
#include <SDL_audio.h>
#include <vorbis/vorbisfile.h>
#include "audio.h"

/* defines */
#define BUFFER_SIZE	((Uint32)256)
#define BUFFER_LIMIT	((Uint32)1000000000)

/* externals */
extern "C"
{
  void warn(char *);
  void error(char *);
}

/* Ogg Vorbis Code */
static size_t RWops_ogg_read(void *ptr, size_t size, size_t nmemb,
			     void *source)
{
  return ((size_t) SDL_RWread((SDL_RWops *) source, ptr, size, nmemb));
}

static int RWops_ogg_seek(void *source, ogg_int64_t offset, int whence)
{
  return (SDL_RWseek((SDL_RWops *) source, offset, whence));
}

static int RWops_ogg_close(void *source)
{
  return 0;
}

static long RWops_ogg_tell(void *source)
{
  return ((long) SDL_RWtell((SDL_RWops *) source));
}

static const ov_callbacks ogg_callbacks = {
  RWops_ogg_read,
  RWops_ogg_seek,
  RWops_ogg_close,
  RWops_ogg_tell
};

/* Audio implemention */
Audio::Audio()
{
  SDL_AudioSpec fmt;

  fmt.freq = 11025;
  fmt.format = AUDIO_S16;
  fmt.channels = 1;
  fmt.samples = BUFFER_SIZE;
  fmt.callback = Audio::_mixAudio;
  fmt.userdata = NULL;
  if (SDL_OpenAudio(&fmt, &_spec) < 0)
    {
      warn(SDL_GetError());
      _initiated = 0;
    }
  else
    {
      SDL_PauseAudio(0);
      _initiated = 1;
    }
}

Audio::~Audio()
{
  SoundNode *node = _soundNode, *next = NULL;

  while (node)
    {
      next = node->next;
      free(node);
      node = next;
    }
  if (_initiated)
    SDL_CloseAudio();
  _instance = NULL;
}

void Audio::_mixAudio(void *unused, Uint8 * stream, int length)
{
  SoundNode *node = NULL, *prev = NULL, *temp = NULL;
  int toWrite = 0;

  if (_mute)
    return;
  prev = NULL;
  node = _soundNode;
  while (node != NULL)
    {
      if (node->position + length > node->sound->length)
	toWrite = node->sound->length - node->position;
      else
	toWrite = length;
      SDL_MixAudio(stream, &node->sound->buffer[node->position], toWrite,
		   SDL_MIX_MAXVOLUME);
      node->position += toWrite;
      if (toWrite < length)
	{
	  if (prev)
	    prev->next = node->next;
	  else
	    _soundNode = node->next;
	  temp = node->next;
	  free(node);
	  node = temp;
	}
      else
	{
	  prev = node;
	  node = node->next;
	}
    }
}

Audio *Audio::instance(void)
{
  if (_instance == NULL)
    _instance = new Audio();
  return _instance;
}

void Audio::setMute(Uint8 mute)
{
  _mute = mute;
}

void Audio::playSound(Sound * sound)
{
  SoundNode *node = NULL;

  if (!sound || !_initiated || _mute)
    return;
  if (sound->single)
    {
      node = _soundNode;
      while (node != NULL)
	{
	  if (node->sound == sound)
	    return;
	  node = node->next;
	}
    }
  node = (SoundNode *) malloc(sizeof(SoundNode));
  node->sound = sound;
  node->position = 0;
  node->next = NULL;
  if (_soundNode != NULL)
    node->next = _soundNode;
  _soundNode = node;
}

Sound *Audio::createSound(Uint8 * data, Uint32 * memcounter)
{
  OggVorbis_File *vf = NULL;
  vorbis_info *vi = NULL;
  SDL_RWops *rwops = NULL;
  Sound *sound = NULL;
  int dummy = 0;
  Uint32 size = 0, bufferSize = 0, bytesRead = 0, counter = 0;

  if (!_initiated)
    return NULL;
  memcpy(&size, data, sizeof(Uint32));
  if (memcounter)
    *memcounter += size + sizeof(Uint32);
  data += sizeof(Uint32);
  rwops = SDL_RWFromMem(data, size);
  vf = (OggVorbis_File *) malloc(sizeof(OggVorbis_File));
  if (ov_open_callbacks(rwops, vf, NULL, 0, ogg_callbacks) != 0)
    return NULL;
  vi = ov_info(vf, -1);
  if (vi->channels != _spec.channels || vi->rate != _spec.freq)
    {
      ov_clear(vf);
      return NULL;
    }
  bufferSize = ov_pcm_total(vf, -1) * _spec.channels * 2;	     /* 2 is for 16-bit sounds */
  if (bufferSize > BUFFER_LIMIT)
    {
      ov_clear(vf);
      return NULL;
    }
  sound = (Sound *) malloc(sizeof(Sound));
  sound->length = bufferSize;
  sound->single = 1;
  sound->buffer = (Uint8 *) malloc(bufferSize);
  while ((bytesRead =
	  ov_read(vf, (char *) &sound->buffer[counter], 4096, 0, 2, 1,
		  &dummy)) != 0)
    {
      counter += bytesRead;
    }
  return sound;
}

void Audio::freeSound(Sound * sound)
{
  SoundNode *node = NULL, *prev = NULL, *temp = NULL;

  if (sound)
    {
      prev = NULL;
      node = _soundNode;
      while (node)
	{
	  if (node->sound == sound)
	    {
	      if (prev)
		prev->next = node->next;
	      else
		_soundNode = node->next;
	      temp = node->next;
	      free(node);
	      node = temp;
	    }
	  else
	    {
	      prev = node;
	      node = node->next;
	    }
	}
      SDL_FreeWAV(sound->buffer);
      free(sound);
    }
}

Audio *Audio::_instance = NULL;
SoundNode *Audio::_soundNode = NULL;
Uint8 Audio::_mute = 0;
