/*
	TuxPuck - board.cpp
	Copyright (C) 2001 Jacob Kroon

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

	Jacob Kroon - d00jkr@efd.lth.se
*/

/* includes */
#include "video.h"
#include "board.h"

/* defines */
#define BOARD_STATE_PLAY        1
#define BOARD_STATE_NEW_PUCK    2
#define BOARD_STATE_CRASH_GLASS	3

/* externals */
extern unsigned char board_png[];

/* Board implemention */
Board::Board()
{
  _state = BOARD_STATE_NEW_PUCK;
  _sdlBoard = Video::instance()->createSurface(board_png, NULL);     /* Load board.png image */
  _pad_1 = new Pad(this, 1);					     /* Create pad 1 */
  _pad_2 = new Pad(this, 2);					     /* Create pad 2 */
  _thePuck = new Puck(this);					     /* Create puck */
  _theGlass = new Glass(this);
  _coeff[0] = 80 * (A1 * A2) / (A1 - A2);
  _coeff[1] = 40 * (A1 + A2) / (A1 - A2);
  _coeff[2] = 80 * (1.0 * W1 / W2) / (1.0 - W1 / W2);
  _coeff[3] = 40 * (1.0 + W1 / W2) / (1.0 - W1 / W2);
}

Board::~Board()
{
  SDL_FreeSurface(_sdlBoard);					     /* Delete board.png surface */
  delete(_pad_1);						     /* Delete pad 1 */
  delete(_pad_2);						     /* Delete pad 2 */
  delete(_thePuck);						     /* Delete puck */
  delete(_theGlass);
}

Pad *Board::getPad(Uint8 who)
{
  if (who == 1)							     /* Return first player */
    return _pad_1;
  else if (who == 2)						     /* Return second player */
    return _pad_2;
  else
    return NULL;						     /* Else return NULL */
}

Puck *Board::getPuck(void)
{
  return _thePuck;
}

void Board::blit(void)
{
  static SDL_Rect rect = { 320 - _sdlBoard->w / 2, 480 - _sdlBoard->h, 0, 0 };	/* Placement of board.png */
  Video::instance()->blit(_sdlBoard, &rect);			     /* Blit it! */
}

Uint8 Board::update(Uint32 time)
{
  static Uint8 turn = 1;
  Uint8 scorer;

  switch (_state)
    {
    case BOARD_STATE_PLAY:
      if (_thePuck->willGetHitBy(_pad_1, time))
	_thePuck->getHitBy(_pad_1, time);
      else if (_thePuck->willGetHitBy(_pad_2, time))
	_thePuck->getHitBy(_pad_2, time);
      _thePuck->move(time);
      break;
    case BOARD_STATE_CRASH_GLASS:
      if (_theGlass->update(time) == 0)
	{
	  _theGlass->erase();
	  _state = BOARD_STATE_NEW_PUCK;
	}
      break;
    case BOARD_STATE_NEW_PUCK:
      if (turn == 1)
	{
	  if (_thePuck->moveTowards(0, -21, time) == 0)
	    {
	      _state = BOARD_STATE_PLAY;
	      turn = 2;
	      _thePuck->setVelocity(0, 0);
	      _thePuck->setState(PUCK_STATE_ALIVE);
	    }
	}
      else
	{
	  if (_thePuck->moveTowards(0, 21, time) == 0)
	    {
	      _state = BOARD_STATE_PLAY;
	      turn = 1;
	      _thePuck->setVelocity(0, 0);
	      _thePuck->setState(PUCK_STATE_ALIVE);
	    }
	}
      break;
    default:
      break;
    }
  if ((scorer = _getScorer()) != 0)
    {
      _state = BOARD_STATE_CRASH_GLASS;
      _theGlass->smash(_thePuck);
      _thePuck->setState(PUCK_STATE_DEAD);
    }
  _pad_1->move(time);
  _pad_2->move(time);
  return scorer;
}

void Board::erase(void)
{
  _pad_1->erase();
  _pad_2->erase();
  _thePuck->erase();
  if (_state == BOARD_STATE_CRASH_GLASS)
    _theGlass->erase();
}

void Board::reblit(void)
{
  float p1, p2, p, g = 0;

  _pad_1->getPosition(NULL, &p1);
  _pad_2->getPosition(NULL, &p2);
  _thePuck->getPosition(NULL, &p);
  if (_state == BOARD_STATE_CRASH_GLASS)
    {
      _theGlass->getPosition(NULL, &g);
      if (g > 0)
	_theGlass->blit();
    }
  if (p < p1)
    {
      _pad_2->blit();						     /* Same for pad 2 */
      _pad_1->blit();						     /* Blit new pad image */
      _thePuck->blit();
    }
  else if (p < p2)
    {
      _pad_2->blit();
      _thePuck->blit();
      _pad_1->blit();
    }
  else
    {
      _thePuck->blit();
      _pad_2->blit();
      _pad_1->blit();
    }
  if (_state == BOARD_STATE_CRASH_GLASS)
    {
      if (g < 0)
	_theGlass->blit();
    }
}

float Board::calcY(float y)
{
  return (_coeff[0] / (_coeff[1] + y));
}

float Board::calcScale(float y)
{
  return _coeff[2] / (_coeff[3] + y);
}

Uint8 Board::_getScorer(void)
{
  float puck_y;

  if (_thePuck->isDead())
    return 0;
  _thePuck->getPosition(NULL, &puck_y);
  if (puck_y > 40)
    return 1;
  else if (puck_y < -40)
    return 2;
  return 0;
}
