/*
	TuxPuck - font.cpp
	Copyright (C) 2001 Jacob Kroon

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

	Jacob Kroon - d00jkr@efd.lth.se
*/

/* includes */
#include <stdlib.h>
#include <string.h>
#include <SDL_endian.h>
#include "video.h"
#include "font.h"

/* Font implemention */
Font::Font(Uint8 * data, Uint32 * memcounter)
{
  Uint32 index = 0, i;
  Uint32 size;
  Uint8 ch;
  Sint16 xSize, ySize;
  SDL_Color palette[256];

  memcpy(&size, data, sizeof(Uint32));
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
  size = SDL_Swap32(size);
#endif
  if (memcounter)
    *memcounter += size + sizeof(Uint32);
  data += sizeof(Uint32);
  for (i = 0; i < 256; i++)
    palette[i].r = palette[i].g = palette[i].b = i;
  _glyph = (Glyph **) calloc(256, sizeof(Glyph *));
  while (index < size)
    {
      ch = data[index++];
      memcpy(&xSize, &data[index], 2);
      index += 2;
      memcpy(&ySize, &data[index], 2);
      index += 2;
      _glyph[ch] = (Glyph *) malloc(sizeof(Glyph));
      memcpy(&_glyph[ch]->xMove, &data[index], 2);
      index += 2;
      memcpy(&_glyph[ch]->yMove, &data[index], 2);
      index += 2;
      memcpy(&_glyph[ch]->advance, &data[index], 2);
      index += 2;
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
      xSize = SDL_Swap16(xSize);
      ySize = SDL_Swap16(ySize);
      _glyph[ch]->xMove = SDL_Swap16(_glyph[ch]->xMove);
      _glyph[ch]->yMove = SDL_Swap16(_glyph[ch]->yMove);
      _glyph[ch]->advance = SDL_Swap16(_glyph[ch]->advance);
#endif
      _glyph[ch]->bitmap =
	SDL_CreateRGBSurfaceFrom(&data[index], xSize, ySize, 8, xSize, 0, 0,
				 0, 0);
      SDL_SetPalette(_glyph[ch]->bitmap, SDL_LOGPAL, palette, 0, 256);
      SDL_SetColorKey(_glyph[ch]->bitmap, SDL_SRCCOLORKEY, 0);
      index += xSize * ySize;
    }
  _x = _y = 0;
}

Font::~Font()
{
  Uint32 i;

  for (i = 0; i < 256; i++)
    if (_glyph[i] != NULL)
      {
	SDL_FreeSurface(_glyph[i]->bitmap);
	free(_glyph[i]);
      }
  free(_glyph);
}

void Font::setColor(Uint8 r, Uint8 g, Uint8 b)
{
  Uint32 i;

  SDL_Color palette[256];
  for (i = 0; i < 256; i++)
    {
      palette[i].r = (Uint8) (r * i / 256.0);
      palette[i].g = (Uint8) (g * i / 256.0);
      palette[i].b = (Uint8) (b * i / 256.0);
    }
  for (i = 0; i < 256; i++)
    if (_glyph[i])
      SDL_SetPalette(_glyph[i]->bitmap, SDL_LOGPAL, palette, 0, 256);

}

void Font::print(char *string, SDL_Rect * fill)
{
  Uint32 ruler = 0;
  Uint32 i;
  SDL_Rect rect, area;

  for (i = 0; i < strlen(string); i++)
    if (_glyph[string[i]])
      {
	rect.x = _x + _glyph[string[i]]->xMove + ruler;
	rect.y = _y - _glyph[string[i]]->yMove;
	rect.w = rect.h = 0;
	Video::instance()->blit(_glyph[string[i]]->bitmap, &rect);
	if (i == 0)
	  area = rect;
	else
	  {
	    if (rect.x < area.x)
	      area.x = rect.x;
	    if (rect.y < area.y)
	      area.y = rect.y;
	    if (rect.x + rect.w > area.x + area.w)
	      area.w = rect.x + rect.w - area.x;
	    if (rect.y + rect.h > area.y + area.h)
	      area.h = rect.y + rect.h - area.y;
	  }
	ruler += _glyph[string[i]]->advance;
      }
  if (fill)
    *fill = area;
}

Uint32 Font::calcWidth(char *string)
{
  Uint32 width = 0;
  Uint32 i;

  if (string == NULL)
    return 0;
  for (i = 0; i < strlen(string); i++)
    width += _glyph[string[i]]->advance;
  return width;
}

void Font::movePen(Uint16 dx, Uint16 dy)
{
  _x += dx;
  _y += dy;
}

void Font::setPen(Uint16 x, Uint16 y)
{
  _x = x;
  _y = y;
}
