/*
	TuxPuck - glass.cpp
	Copyright (C) 2001 Jacob Kroon

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

	Jacob Kroon - d00jkr@efd.lth.se
*/

/* includes */
#include "video.h"
#include "board.h"
#include "glass.h"

/* defines */
#define GRAVITY			((float)-0.001)

/* externals */
extern unsigned char glass_png[];
extern unsigned char piece1_png[];
extern unsigned char piece2_png[];
extern unsigned char piece3_png[];
extern unsigned char piece4_png[];
extern unsigned char piece5_png[];
extern unsigned char crash_ogg[];

/* Glass implemention */
Glass::Glass(Board * board)
{
  Uint8 i;

  _theBoard = board;
  _x = _z = 0;
  _sdlGlass = Video::instance()->createSurface(glass_png, NULL);
  for (i = 0; i < 5; i++)
    _piece[i] = new GlassPiece(board, this, i);
  _oldRect.x = _oldRect.y = _oldRect.w = _oldRect.h = 0;
  _sndCrash = Audio::instance()->createSound(crash_ogg, NULL);
}

Glass::~Glass()
{
  Uint8 i = 0;

  SDL_FreeSurface(_sdlGlass);
  for (i = 0; i < 5; i++)
    delete(_piece[i]);
  Audio::instance()->freeSound(_sndCrash);
}

void Glass::smash(Puck * puck)
{
  float x, z;
  Uint8 i;

  puck->getPosition(&x, &z);
  _x = x - 1.2;
  if (z < 0)
    _z = -40;
  else
    _z = 40;
  for (i = 0; i < 5; i++)
    _piece[i]->reset();
  Audio::instance()->playSound(_sndCrash);
}

Uint8 Glass::update(Uint32 time)
{
  Uint8 done = 1;
  Uint8 i;

  for (i = 0; i < 5; i++)
    if (_piece[i]->update(time) != 0)
      done = 0;
  if (done)
    return 0;
  return 1;
}

void Glass::getPosition(float *x, float *z)
{
  if (x)
    *x = _x;
  if (z)
    *z = _z;
}

void Glass::erase(void)
{
  Uint8 i;

  Video::instance()->erase(&_oldRect);
  for (i = 0; i < 5; i++)
    _piece[i]->erase();
}

void Glass::blit(void)
{
  SDL_Surface *scaledSurface = NULL;
  float scale, y;
  Uint8 i;

  scale = _theBoard->calcScale(_z);
  y = _theBoard->calcY(_z);
  scaledSurface = Video::instance()->scaleSurface(_sdlGlass, scale);
  _oldRect.x = (Sint16) (320 - scaledSurface->w / 2 + scale * _x * 12.7);
  _oldRect.y = (Sint16) (y - scaledSurface->h);
  Video::instance()->blit(scaledSurface, &_oldRect);
  SDL_FreeSurface(scaledSurface);
  for (i = 0; i < 5; i++)
    _piece[i]->blit();
}

/* GlassPiece implemention */
GlassPiece::GlassPiece(Board * board, Glass * glass, Uint32 id)
{
  _theBoard = board;
  _x = _y = _z = _dx = _dy = 0;
  _id = id;
  _oldRect.x = _oldRect.y = _oldRect.w = _oldRect.h = 0;
  _theGlass = glass;
  switch (id)
    {
    case 0:
      _sdlImage = Video::instance()->createSurface(piece1_png, NULL);
      break;
    case 1:
      _sdlImage = Video::instance()->createSurface(piece2_png, NULL);
      break;
    case 2:
      _sdlImage = Video::instance()->createSurface(piece3_png, NULL);
      break;
    case 3:
      _sdlImage = Video::instance()->createSurface(piece4_png, NULL);
      break;
    case 4:
      _sdlImage = Video::instance()->createSurface(piece5_png, NULL);
      break;
    default:
      break;
    }
  if (id == 1)							     /* seems piece1.png doesnt have 0 as black */
    SDL_SetColorKey(_sdlImage, SDL_SRCCOLORKEY, 1);
}

GlassPiece::~GlassPiece()
{
  SDL_FreeSurface(_sdlImage);
}

void GlassPiece::erase(void)
{
  Video::instance()->erase(&_oldRect);
}

void GlassPiece::blit(void)
{
  SDL_Surface *scaledSurface = NULL;
  float scale, y;

  scale = _theBoard->calcScale(_z);
  y = _theBoard->calcY(_z) - _y * scale;
  scaledSurface = Video::instance()->scaleSurface(_sdlImage, scale);
  _oldRect.x = (Sint16) (320 + scale * _x * 12.7);
  _oldRect.y = (Sint16) (y - scaledSurface->h);
  Video::instance()->blit(scaledSurface, &_oldRect);
  SDL_FreeSurface(scaledSurface);
}

Uint8 GlassPiece::update(Uint32 time)
{
  _dy += GRAVITY * time;
  _x += _dx * time;
  _y += _dy * time;
  if (_y < -10)
    return 0;
  return 1;
}

void GlassPiece::reset(void)
{
  float x, z;
  float dx;

  _theBoard->getPuck()->getVelocity(&dx, NULL);
  _theGlass->getPosition(&x, &z);
  switch (_id)
    {
    case 0:
      _x = x - 2.5;
      _y = 0;
      _z = z;
      _dx = -0.01;
      _dy = 0.25;
      break;
    case 1:
      _x = x - 1.9;
      _y = 0.2;
      _z = z;
      _dx = -0.008;
      _dy = 0.28;
      break;
    case 2:
      _x = x + 0.5;
      _y = 0;
      _z = z;
      _dx = 0.001;
      _dy = 0.3;
      break;
    case 3:
      _x = x + 1.0;
      _y = 0.7;
      _z = z;
      _dx = 0.007;
      _dy = 0.27;
      break;
    case 4:
      _x = x + 1.1;
      _y = 0.1;
      _z = z;
      _dx = 0.012;
      _dy = 0.26;
      break;
    default:
      break;
    }
  _dx += dx * 0.1;
}
