/*
	TuxPuck - player.cpp
	Copyright (C) 2001 Jacob Kroon

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

	Jacob Kroon - d00jkr@efd.lth.se
*/

/* includes */
#include <math.h>
#include <SDL_mouse.h>
#include "video.h"
#include "board.h"
#include "player.h"

/* defines */
#define MIN_MOUSE_SPEED		((float)0.05)
#define MAX_MOUSE_SPEED		((float)0.50)

/* Player implemention */
Player::Player(Pad * pad, char *name)
{
  _name = name;
  _thePad = pad;
  _points = 0;
  _speed = (MIN_MOUSE_SPEED + MAX_MOUSE_SPEED) / 2.0;;
}

Player::~Player()
{
}

void Player::givePoint(void)
{
  _points++;
}

Uint8 Player::getPoints(void)
{
  return _points;
}

char *Player::getName(void)
{
  return _name;
}

/* HumanPlayer implemention */
HumanPlayer::HumanPlayer(Pad * pad, char *name):Player(pad, name)
{
}

HumanPlayer::~HumanPlayer()
{
}

void HumanPlayer::setSpeed(Uint8 speed)
{
  _speed =
    MIN_MOUSE_SPEED + (MAX_MOUSE_SPEED - MIN_MOUSE_SPEED) * speed / 10.0;
}

void HumanPlayer::update(Uint32 time)
{
  int dx, dy;

  SDL_GetRelativeMouseState(&dx, &dy);
  if (time != 0)
    _thePad->setVelocity((float) dx / time * _speed,
			 (float) -dy / time * _speed);
  else
    _thePad->setVelocity(0, 0);
}

/* AIPlayer implemention */
AIPlayer::AIPlayer(Pad * pad, Puck * puck, char *name):Player(pad, name)
{
  _thePuck = puck;
  _hitPower = 0.0;
  _state = PLAYER_STATE_IDLE;
}

AIPlayer::~AIPlayer()
{
}

void AIPlayer::update(Uint32 time)
{
  static float dx = 0, dy = 0;
  static float counter = 3.14 / 2;
  float hit_speed;
  float puck_x, puck_y;
  float pad_x, pad_y;

  _thePuck->getPosition(&puck_x, &puck_y);
  _thePad->getPosition(&pad_x, &pad_y);
  if (!
      (_state == PLAYER_STATE_WIN_POINT || _state == PLAYER_STATE_LOOSE_POINT
       || _state == PLAYER_STATE_WIN_GAME
       || _state == PLAYER_STATE_LOOSE_GAME))
    {

      if (_thePuck->isDead())
	this->setState(PLAYER_STATE_IDLE);
      else if (puck_y > 20)
	{
	  if (_state != PLAYER_STATE_AIM && _state != PLAYER_STATE_HIT)
	    this->setState(PLAYER_STATE_BACKUP);
	}
      else
	this->setState(PLAYER_STATE_IDLE);
    }
  counter += 0.005 * time;
  switch (_state)
    {
    case PLAYER_STATE_WIN_POINT:
    case PLAYER_STATE_LOOSE_POINT:
    case PLAYER_STATE_IDLE:
      dx = (sin(counter / 2.0) * 0.03 + (puck_x - pad_x) * 0.005) * _speed;
      dy = (sin(counter / 4.0) * 0.01 + (35 - pad_y) * 0.005) * _speed;
      break;
    case PLAYER_STATE_BACKUP:
      dx = dy = 0;
      if (_thePad->moveTowards(pad_x, 38, (Uint32) (time * _speed)) == 0)
	{
	  this->setState(PLAYER_STATE_AIM);
	}
      break;
    case PLAYER_STATE_AIM:
      dx = (puck_x - pad_x);
      dy = (puck_y - pad_y);
      if (dy > -0.4)
	{							     /* prevent player from getting stuck */
	  this->setState(PLAYER_STATE_BACKUP);
	  break;
	}
      hit_speed = sqrt(dx * dx + dy * dy);
      dx = dx / hit_speed * _hitPower * 0.1;
      dy = dy / hit_speed * _hitPower * 0.1;
      this->setState(PLAYER_STATE_HIT);
      break;
    case PLAYER_STATE_HIT:
      if (pad_y < 21)
	this->setState(PLAYER_STATE_IDLE);
      break;
    case PLAYER_STATE_WIN_GAME:
    case PLAYER_STATE_LOOSE_GAME:
      dx = dy = 0;
      break;
    default:
      break;
    }
  _thePad->setVelocity(dx, dy);
}
