/*
	TuxPuck - tuxpuck.cpp
	Copyright (C) 2001 Jacob Kroon

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

	Jacob Kroon - d00jkr@efd.lth.se
*/

/* includes */
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <SDL.h>
#include "scoreboard.h"
#include "video.h"
#include "tuxpuck.h"
#include "board.h"
#include "player.h"
#include "aiplayers.h"
#include "font.h"
#include "menu.h"

/* externals */
extern unsigned char sympa_fnt[];
extern unsigned char vulgar_fnt[];
extern "C"
{
  void error(char *);
  void warn(char *);
}

/* TuxPuck implemention */
TuxPuck::TuxPuck()
{
  if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO) < 0)		     /* Init SDL */
    error(SDL_GetError());
  Audio::instance();
  Video::instance();
  _fntMenu = new Font(vulgar_fnt, NULL);
  _fntFps = new Font(sympa_fnt, NULL);
  _gameMenu = new GameMenu(_fntMenu, _fntFps);
  _opponentMenu = new OpponentMenu(_fntMenu);
  Video::instance()->save();
  srand(time(NULL));
  _settings = (Settings *) calloc(1, sizeof(Settings));
  _settings->sound = 1;
  _settings->fullscreen = 0;
  _settings->transparency = 1;
  _settings->mouseSpeed = 5;
  _settings->showFps = 0;
#ifdef windows
#define SETTINGSFILE "tuxpuck.ini"
#else
  char *homeDir = NULL;
  homeDir = getenv("HOME");
  sprintf(_settingsFile, "%s/.tuxpuckrc", homeDir);
#define SETTINGSFILE _settingsFile
#endif
  _readSettings();
  Audio::instance()->setMute(!_settings->sound);
  if (_settings->fullscreen)
    Video::instance()->toggleFullscreen();
}

TuxPuck::~TuxPuck()
{
  delete(_opponentMenu);
  delete(_gameMenu);
  delete(_fntMenu);
  delete(_fntFps);
  delete(Audio::instance());
  delete(Video::instance());
  SDL_Quit();
  _saveSettings();
  free(_settings);
}

void TuxPuck::start(void)
{
  Uint8 choice = 0, loop = 1, opponent = 0;

  while (loop)
    {
      choice = _gameMenu->getCommand();
      switch (choice)
	{
	case GAMEMENU_START_GAME:
	  opponent = _opponentMenu->getCommand();
	  if (opponent != OPPONENTMENU_CANCEL)
	    _playMatch(opponent);
	  break;
	case GAMEMENU_QUIT:
	  loop = 0;
	  break;
	default:
	  break;
	}
    }
}

void TuxPuck::_playMatch(Uint8 opponent)
{
  SDL_Event event;
  SDL_Rect fpsRect = { 0, 0, 0, 0 };
  Uint8 loop = 1, scorer = 0;
  Uint32 elapsedTime = 0, timeStamp = 0, timeCounter = 0, frameCounter = 0;
  char buffer[50];
  HumanPlayer *p1 = NULL;
  AIPlayer *p2 = NULL;
  Board *board = NULL;
  Scoreboard *scoreboard = NULL;
  IngameMenu *ingameMenu = NULL;

  memset(buffer, 0, 50);
  _fntFps->setPen(577, 18);
  ingameMenu = new IngameMenu(_fntMenu);
  board = new Board();
  scoreboard = new Scoreboard();
  p1 = new HumanPlayer(board->getPad(1), "Human");
  switch (opponent)
    {
    case OPPONENTMENU_TUX:
      p2 = new TuxPlayer(board->getPad(2), board->getPuck());
      break;
    case OPPONENTMENU_CARTMAN:
      p2 = new CartmanPlayer(board->getPad(2), board->getPuck());
      break;
    case OPPONENTMENU_HOMER:
      p2 = new HomerPlayer(board->getPad(2), board->getPuck());
      break;
    default:
      break;
    }
  board->blit();						     /* Blit it! */
  scoreboard->blit();
  ((TuxPlayer *) p2)->blit();
  Video::instance()->update();
  Video::instance()->save();					     /* Save background to use for erasing */
  timeStamp = SDL_GetTicks();
  SDL_PumpEvents();
  SDL_GetRelativeMouseState(NULL, NULL);
#ifndef _DEBUG
  SDL_WM_GrabInput(SDL_GRAB_ON);
#endif
  if (!_settings->transparency)
    {
      board->getPad(1)->setTransparency(0);
      board->getPad(2)->setTransparency(0);
    }
  p1->setSpeed(_settings->mouseSpeed);
  while (loop)
    {
      while (SDL_PollEvent(&event))
	switch (event.type)
	  {
	  case SDL_KEYDOWN:
	    switch (event.key.keysym.sym)
	      {
	      case SDLK_ESCAPE:
		switch (ingameMenu->getCommand())
		  {
		  case INGAMEMENU_EXIT_GAME:
		    loop = 0;
		    break;
		  case INGAMEMENU_CONTINUE:
		    break;
		  default:
		    break;
		  }
		timeStamp = SDL_GetTicks();
		break;
	      case SDLK_F1:
		if (_settings->sound)
		  {
		    _settings->sound = 0;
		    Audio::instance()->setMute(1);
		  }
		else
		  {
		    _settings->sound = 1;
		    Audio::instance()->setMute(0);
		  }
		break;
	      case SDLK_F2:
		if (_settings->transparency)
		  _settings->transparency = 0;
		else
		  _settings->transparency = 1;
		board->getPad(1)->setTransparency(_settings->transparency);
		board->getPad(2)->setTransparency(_settings->transparency);
		break;
	      case SDLK_F3:
		break;
	      case SDLK_F4:
		if (_settings->showFps)
		  {
		    _settings->showFps = 0;
		    Video::instance()->erase(&fpsRect);
		  }
		else
		  _settings->showFps = 1;
		break;
	      case SDLK_F5:
		if (_settings->mouseSpeed > 1)
		  _settings->mouseSpeed--;
		p1->setSpeed(_settings->mouseSpeed);
		break;
	      case SDLK_F6:
		if (_settings->mouseSpeed < 10)
		  _settings->mouseSpeed++;
		p1->setSpeed(_settings->mouseSpeed);
		break;
	      case SDLK_f:
		if (_settings->fullscreen)
		  _settings->fullscreen = 0;
		else
		  _settings->fullscreen = 1;
		Video::instance()->toggleFullscreen();
		break;
	      default:
		break;
	      }
	    break;
	  case SDL_QUIT:
	    loop = 0;
	    break;
	  }
      SDL_Delay(5);						     /* needed for _to_ fast video cards */
      elapsedTime = SDL_GetTicks() - timeStamp;
      timeStamp = SDL_GetTicks();
      p1->update(elapsedTime);
      p2->update(elapsedTime);
      scoreboard->update(elapsedTime);
      if ((scorer = board->update(elapsedTime)) != 0)
	{
	  scoreboard->addPoint(scorer);
	  if (scorer == 1)
	    {
	      p1->givePoint();
	      ((AIPlayer *) p2)->setState(PLAYER_STATE_LOOSE_POINT);
	    }
	  else
	    {
	      p2->givePoint();
	      ((AIPlayer *) p2)->setState(PLAYER_STATE_WIN_POINT);
	    }
	  if (p1->getPoints() >= 15 || p2->getPoints() >= 15)
	    {
	      if (p1->getPoints() == 15)
		p2->setState(PLAYER_STATE_LOOSE_GAME);
	      else
		p2->setState(PLAYER_STATE_WIN_GAME);
	    }
	}
      board->erase();
      scoreboard->reblit();
      if (p2->reblit(elapsedTime) == 0)
	loop = 0;
      board->reblit();						     /* Redraw board things */
      if (_settings->showFps)
	{
	  Video::instance()->erase(&fpsRect);
	  _fntFps->print(buffer, &fpsRect);
	  frameCounter++;
	  timeCounter += elapsedTime;
	  if (timeCounter > 200 || frameCounter > 200)
	    {
	      sprintf(buffer, "FPS:%3.0f",
		      (float) frameCounter / timeCounter * 1000.0);
	      frameCounter = 0;
	      timeCounter = 0;
	    }
	}
      Video::instance()->update();
    }
#ifndef _DEBUG
  SDL_WM_GrabInput(SDL_GRAB_OFF);
#endif
  delete(p1);
  delete(p2);
  delete(board);
  delete(scoreboard);
  delete(ingameMenu);
}

void TuxPuck::_readSettings(void)
{
  FILE *file = NULL;
  char buffer[100], buffer2[100];
  Uint32 uint32 = 0;

  if ((file = fopen(SETTINGSFILE, "r")) == NULL)
    if ((file = fopen(SETTINGSFILE, "w")) == NULL)
      {
	warn("Couldn't open settingsfile");
	return;
      }
  while (fgets(buffer, 100, file) != 0)
    {
      if (sscanf(buffer, "%s %d\n", buffer2, &uint32) != 2)
	{
	  warn("Corrupt settingsfile");
	  fclose(file);
	  return;
	}
      else if (strcmp(buffer2, "SOUND") == 0)
	_settings->sound = (Uint8) uint32;
      else if (strcmp(buffer2, "FULLSCREEN") == 0)
	_settings->fullscreen = (Uint8) uint32;
      else if (strcmp(buffer2, "TRANSPARENCY") == 0)
	_settings->transparency = (Uint8) uint32;
      else if (strcmp(buffer2, "MOUSESPEED") == 0)
	_settings->mouseSpeed = (Uint8) uint32;
      else if (strcmp(buffer2, "SHOWFPS") == 0)
	_settings->showFps = (Uint8) uint32;
      else
	{
	  warn("Ignoring parameter:");
	  warn(buffer2);
	}
    }
  fclose(file);
}

void TuxPuck::_saveSettings(void)
{
  FILE *file = NULL;

  if ((file = fopen(SETTINGSFILE, "w")) == NULL)
    {
      warn("Couldn't write settingsfile");
      return;
    }
  fprintf(file, "SOUND %d\n", _settings->sound);
  fprintf(file, "FULLSCREEN %d\n", _settings->fullscreen);
  fprintf(file, "TRANSPARENCY %d\n", _settings->transparency);
  fprintf(file, "MOUSESPEED %d\n", _settings->mouseSpeed);
  fprintf(file, "SHOWFPS %d\n", _settings->showFps);
  fclose(file);
}

/* main function */
int main(int argc, char **argv)
{
  TuxPuck *tuxPuck = NULL;

  tuxPuck = new TuxPuck();
  tuxPuck->start();
  delete(tuxPuck);
  return 0;
}
